import collections
import gensim.corpora as corpora
import nltk
import pandas as pd
from bertopic import BERTopic
from gensim.models.coherencemodel import CoherenceModel
from octis.evaluation_metrics.coherence_metrics import Coherence
from octis.evaluation_metrics.diversity_metrics import TopicDiversity
from sentence_transformers import SentenceTransformer, models
from typing import List, Dict, Tuple, Union, Literal

nltk.download('stopwords')

nltk.download('punkt')


def initialize_model(embedding_model: Union[SentenceTransformer, str] = "specter",
                     nr_topics: int = None,
                     n_gram_range: Union[Literal["unigram", "bigram"], Tuple[int, int]] = "unigram",
                     verbose="False"):
    """
        Conducts type checks on the input variables and converts certain parameters to the proper types dependent on their input.
        Not intended to be called by the user, but instead used internally by the fit function.


        Parameters
        ----------
        embedding_model
                Either a SentenceTransformer, or a string with values "specter". "aspire", or "scibert".
        nr_topics: int
                The desired number of topics, if not specified the results will be determined by HDBSCAN's reduction step.
        n_gram_range
                String indicating whether the user would like unigram or bigram or tuple of ints.
        verbose
                Boolean indicating verbose output.

        Returns
        -------
        embedding_model
                SentenceTransformer language model
        n_gram_range
                tuple indicating the level of n_gram
            """

    if not isinstance(nr_topics, int) and nr_topics is not None:
        raise TypeError("nr_topics must be an int")
    if not isinstance(verbose, bool) and nr_topics is not None:
        raise TypeError("verbose must be a bool")

    if isinstance(n_gram_range, str):
        if n_gram_range == "unigram":
            n_gram_range = (1, 1)
        elif n_gram_range == "bigram":
            n_gram_range = (2, 2)
        else:
            raise AttributeError("n_gram_type must equal \"unigram\" or \"bigram\" if it is a string")
    elif isinstance(n_gram_range, tuple):
        pass
    else:
        raise TypeError("n_gram_range must be an string or tuple")

    if isinstance(embedding_model, str):
        if embedding_model.lower() == "specter":
            embedding_model = SentenceTransformer('allenai-specter')

        elif embedding_model.lower() == "aspire":
            word_embedding_model = models.Transformer('allenai/aspire-sentence-embedder', max_seq_length=512)
            pooling_model = models.Pooling(word_embedding_model.get_word_embedding_dimension(), pooling_mode='cls')
            embedding_model = SentenceTransformer(modules=[word_embedding_model, pooling_model])

        elif embedding_model.lower() == "scibert":
            word_embedding_model = models.Transformer('allenai/scibert_scivocab_uncased', max_seq_length=512)
            pooling_model = models.Pooling(word_embedding_model.get_word_embedding_dimension(), pooling_mode='cls')
            embedding_model = SentenceTransformer(modules=[word_embedding_model, pooling_model])

        else:
            raise AttributeError("embedding model should either be a string specifying one of the 3 pre-loaded "
                                 "models, or the desired language model")
    embedding_model = embedding_model

    n_gram_range = n_gram_range

    return embedding_model, n_gram_range


def fit(data: List[str],
        embedding_model: Union[SentenceTransformer, str] = "specter",
        nr_topics: int = None,
        n_gram_range: Union[Literal["unigram", "bigram"], Tuple[int, int]] = "unigram",
        verbose: bool = False):
    """
        Fits a BERTopic model on the data. After fitting the topic assigned to each document is stored
        in the 'topics' attribute, the coherence and diversity measures are stored in the
        'coherence' and 'diversity' attributes respectively, and the amount of documents assigned to each topic
        are stored in the 'topic_sizes' attribute.


        Parameters
        ----------
        data
                List of the documents
        embedding_model
                Either a SentenceTransformer, or a string with values "specter". "aspire", or "scibert".
        nr_topics
                The desired number of topics, if not specified the results will be determined by HDBSCAN's reduction step.
        n_gram_range
                String indicating whether the user would like unigram or bigram, or tuple of ints.
        verbose
                Boolean indicating verbose output.

        Returns
        -------
        topics
                a list of integers representing the topic the corresponding document was assigned to
        probabilities
                a
        metrics
                a dictionary of the 2 metrics with keys "Coherence" and "Diversity"
        topic_sizes
                a dictionary with key: topic number and the value: the number of documents assigned the said topic
        topic_model
                the fitted BERTopic model
        topic_words
                a dictionary with key: the topic number and the value: a list of strings
        """
    opts = dict()

    if not isinstance(embedding_model, SentenceTransformer) and not isinstance(n_gram_range, tuple):
        embedding_model, opts['n_gram_range'] = initialize_model(embedding_model, nr_topics, n_gram_range, verbose)

    if not isinstance(data, list) or (isinstance(data, list) and not isinstance(data[0], str)):
        raise TypeError("Data must be a list of strings")

    topic_model = BERTopic(embedding_model=embedding_model,
                           nr_topics=nr_topics,
                           verbose=verbose,
                           **opts)
    topics, probabilities = topic_model.fit_transform(data)
    # metrics = _calculate_metrics(data, topic_model, topics, **opts)
    topic_sizes = _calculate_topic_sizes(topics)
    topic_words = topic_model.topic_representations_

    return topics, probabilities, topic_sizes, topic_model, topic_words


def _calculate_topic_sizes(topics: List[int]):
    """
    counts the number of documents assigned to each topic and stores the topic sizes in the 'topic_sizes'
    attribute in the form of a dict.

    Parameters
    ----------
    topics
        List of topic assignments for each document

    Returns
    -------
    Sets the topic size parameter to a collection object containing the topics and their sizes.

    """

    counter = collections.Counter(topics)
    topic_sizes = dict(counter)
    return topic_sizes


def calculate_metrics(texts: List[str], topic_model: BERTopic, topics: List[int], n_gram_range: Tuple[int, int] = None):
    """
    Calculates the Topic Coherence and Topic Diversity of the topic model.
    Parameters
    ----------
    texts
         The list of documents

    Returns
    -------
    dict
        A dict containing the metrics as keys, and their respective scores as values.
    """

    topic_words = {}
    topic_dict = topic_model.topic_representations_
    for k in topic_dict.keys():
        topic_words[k] = [x[0] for x in topic_dict[k]]
    word_list = list(topic_words.values())
    word_list.pop(0)

    topic_diversity = TopicDiversity(topk=10)

    coherence_score = _calculate_coherence(topic_model, texts, topics)

    output_tm = {"topics": word_list}
    diversity_score = topic_diversity.score(output_tm)

    return {"Coherence": coherence_score, "Diversity": diversity_score}


def _calculate_coherence(topic_model: BERTopic, docs: List[str], topics):
    """
    Internal method for calculating the coherence when the n_gram_range is set to something other than (1,1)
    Parameters
    ----------
    topic_model
        The BERTopic object
    docs
        The list of documents

    Returns
    -------
        The coherence score

    """

    # Preprocess Documents
    documents = pd.DataFrame({"Document": docs,
                              "ID": range(len(docs)),
                              "Topic": topics})
    documents_per_topic = documents.groupby(['Topic'], as_index=False).agg({'Document': ' '.join})
    cleaned_docs = topic_model._preprocess_text(documents_per_topic.Document.values)

    # Extract vectorizer and analyzer from BERTopic
    vectorizer = topic_model.vectorizer_model
    analyzer = vectorizer.build_analyzer()

    # Extract features for Topic Coherence evaluation
    tokens = [analyzer(doc) for doc in cleaned_docs]
    dictionary = corpora.Dictionary(tokens)
    corpus = [dictionary.doc2bow(token) for token in tokens]
    topic_words = [[words for words, _ in topic_model.get_topic(topic)]
                   for topic in range(len(set(topics)) - 1)]

    # Evaluate
    coherence_model = CoherenceModel(topics=topic_words,
                                     texts=tokens,
                                     corpus=corpus,
                                     dictionary=dictionary,
                                     processes=1,
                                     coherence='c_v')
    coherence = coherence_model.get_coherence()
    return coherence


def create_barcharts(topics, topic_model: BERTopic, path=""):
    """
    creates and saves the BERTopic barcharts

    Parameters
    ----------
    topics
            the list of topic assignments output from the fit function
    topic_model
            fitted BERTopic model output from the fit function
    path
        A string containing the desired path to save the barchart

    Returns
    -------
    Barcharts are saved in .html and .png format in the desired directory.

    """

    if not isinstance(path, str):
        raise TypeError("path must be a string")

    if path != "":
        path = path + 'barchart'

    # model_path = base_path + iter_version + beamline + "_" + model_use + "_" + n_gram_type + "/"
    fig = topic_model.visualize_barchart(top_n_topics=len(topics))
    fig_name = path + ".html"
    fig_name_png = path + ".png"
    if path != "":
        fig.write_html(fig_name)
        fig.write_image(fig_name_png)
    return fig



from berteley import preprocessing
import pytest


def test_combine_hyphens():
    # deprecated
    string1 = "-this is A test string"
    string2 = "this is A test string-"
    string3 = "this-is A test string"
    string4 = "this is - A test string"
    string5 = " - this is A test string"
    string6 = "this is A test string - "
    string7 = "this is A- test string"
    string8 = "this is -A test string"
    hyphen_string = "this is A test string"

    assert preprocessing.combine_hyphens(string1) == hyphen_string
    assert preprocessing.combine_hyphens(string2) == hyphen_string
    assert preprocessing.combine_hyphens(string3) == 'thisis A test string'
    assert preprocessing.combine_hyphens(string4) == hyphen_string
    assert preprocessing.combine_hyphens(string5) == hyphen_string
    assert preprocessing.combine_hyphens(string6) == hyphen_string
    assert preprocessing.combine_hyphens(string7) == hyphen_string
    assert preprocessing.combine_hyphens(string8) == hyphen_string


def test_remove_punctuation():
    string1 = "I'm a test string!"
    string2 = "I,m a ! test_ string"
    assert preprocessing.remove_punctuation(string1) == "Im a test string"
    assert preprocessing.remove_punctuation(string2) == "Im a  test string"

    string3 = 'x-ray'
    assert preprocessing.remove_punctuation(string3) == "xray"


def test_lemmatize():
    string1 = "there are a lot of worlds in the galaxy"
    assert preprocessing.lemmatize(string1) == 'there be a lot of world in the galaxy'


def test_remove_stopwords():
    string1 = "iii ion light advanced words ab"
    assert preprocessing.remove_stopwords(string1, allow_abbrev=True) == 'ion words ab'
    assert preprocessing.remove_stopwords(string1, allow_abbrev=False) == 'ion words'


def test_input_errors():
    with pytest.raises(TypeError):
        preprocessing.combine_hyphens(123)
    with pytest.raises(AttributeError):
        preprocessing.remove_punct_df(123)
    with pytest.raises(TypeError):
        preprocessing.lemmatize(123)
    with pytest.raises(TypeError):
        preprocessing.remove_stopwords(123)
    with pytest.raises(TypeError):
        preprocessing.remove_stopwords("abc", "True")


def test_expand_contractions():
    string1 = "I\'ll be there within 5 minutes"
    assert preprocessing.expand_contractions(string1) == "I will be there within 5 minutes"


def test_remove_extraspace():
    string1 = "\n\ni enjoy   going to the movies   with my friends\n"
    assert preprocessing.remove_extraspace(string1) == "i enjoy going to the movies with my friends"


def test_remove_html():
    string1 = "\n\ni enjoy <p> going to the movies </p> with my friends\n"
    assert preprocessing.remove_html(string1) == "\n\ni enjoy   going to the movies   with my friends\n"


def test_preprocess():
    string1 = "Good afternoon, I'm here for my 12:30 x-ray appointment with Dr. Ushizima. I was wondering if it would be possible to get a follow up appointment?"
    docs = [string1]

    assert preprocessing.preprocess(docs) == ["good afternoon xray appointment dr ushizima wonder would possible get follow appointment"]


def test_preprocess_parallel():
    string1 = "The following setup uses a large encoder followed by a simple decoding stage. This converts the input text into a vector."
    string2 = "At the Advanced Light Source we use a lot of equipment"
    assert preprocessing.preprocess_parallel([string1, string2], allow_abbrev=False, n_workers=2) == ["follow setup large encoder follow simple decoding stage convert input text vector"]



import pytest
import os
from sklearn.datasets import fetch_20newsgroups
import numpy as np

from berteley.models import initialize_model, fit, create_barcharts, calculate_metrics

# gets rid of extraneous warnings
os.environ["TOKENIZERS_PARALLELISM"] = "false"


@pytest.fixture
def data():
    return fetch_20newsgroups(subset='all', remove=('headers', 'footers', 'quotes'))['data'][:100]


@pytest.fixture
def test(data):
    model = "specter"
    topics, probabilities, topic_sizes, topic_model, topic_words = fit(data, embedding_model=model,
                                                                       n_gram_range="unigram", verbose=True)
    return {"topics": topics, "probs": probabilities, "topic_sizes": topic_sizes, "topic_model": topic_model,
            "topic_words": topic_words}


def test_berteley_init():
    # modelPath = "../bert-base-nli-mean-tokens"
    with pytest.raises(AttributeError):
        initialize_model(embedding_model="default")

    with pytest.raises(TypeError):
        initialize_model(nr_topics='2')

    with pytest.raises(TypeError):
        initialize_model(n_gram_range=2)

    with pytest.raises(AttributeError):
        initialize_model(n_gram_range='2')

    initialize_model(embedding_model="specter", n_gram_range="unigram")
    initialize_model(embedding_model="aspire", n_gram_range="bigram")
    initialize_model(embedding_model="scibert", n_gram_range=(1, 1))
    initialize_model(embedding_model="specter", n_gram_range=(2, 2))


def test_fit_input_type(data):
    df = {"data": data}
    with pytest.raises(TypeError):
        fit(df, embedding_model="specter", n_gram_type="bigram", verbose=True)


def test_fit_attributes(test, data):
    assert isinstance(test["topic_sizes"], dict)

    assert isinstance(test["topic_words"], dict)

    metrics = calculate_metrics(data, test["topic_model"], test["topics"])
    assert isinstance(metrics["Coherence"], np.float64)
    assert isinstance(metrics["Diversity"], float)


def test_bigram(data):
    topics, probabilities, topic_sizes, topic_model, topic_words = fit(data,
                                                                       embedding_model="specter",
                                                                       n_gram_range="unigram",
                                                                       verbose=True)

    metrics = calculate_metrics(data, topic_model, topics)
    assert isinstance(metrics["Coherence"], np.float64)
    assert isinstance(metrics["Diversity"], float)


def test_figures(test, tmp_path):
    # os.remove(path + "barchart.html")
    # os.remove(path + "barchart.png")
    # test["topic_model"].visualize_barchart(path=tmp_path)
    create_barcharts(test["topics"], test["topic_model"], path=str(tmp_path) + "/")
    # self.assertTrue(os.path.exists(path + "barchart.html"))
    # self.assertTrue(os.path.exists(path + "barchart.png"))
    assert os.path.exists(str(tmp_path) + "/" + "barchart.html")
    assert os.path.exists(str(tmp_path) + "/" + "barchart.png")
    with pytest.raises(TypeError):
        test["topic_model"].visualize_barchart(123)

    # self.assertRaises(TypeError, test.create_barcharts, 123)



import contextlib
import re
import string
from functools import partial
import spacy
import numpy as np
from nltk.tokenize import word_tokenize
import nltk
from nltk.corpus import stopwords
from bs4 import BeautifulSoup
import contractions
from joblib import delayed, Parallel, parallel
import csv
from pathlib import Path
import subprocess
from alive_progress import alive_it, alive_bar

from typing import List

nltk.download('stopwords')
try:
    nlp = spacy.load('en_core_web_lg')
except OSError:
    subprocess.run(["python", "-m", "spacy", "download", "en_core_web_lg"])
    nlp = spacy.load('en_core_web_lg')

here = Path(__file__).parent
file = open(here / "berteley_stopwords.csv", "r")

data = list(csv.reader(file, delimiter=","))

STOPWORDS = [x[0] for x in data]
NLTK_STOPWORDS = set(stopwords.words('english'))


def combine_hyphens(doc: str):
    """
    Removes hyphens and concatenates all hyphenated words together

    Parameters
    ----------
    doc
        A string for a single document

    Returns
    -------
        A string with hyphenated words combined


    Examples
    x-ray -> xray


    """

    if not isinstance(doc, str):
        raise TypeError("row_string must be a string")

    row_string = re.sub("-", "", doc)
    return " ".join(row_string.split())
    # return re.sub("(?<=\S)-(?=\S)", "", row_string)


def remove_punctuation(doc: str):
    """
    Removes all punctuation from a string using the punctuation list in the string library

    Parameters
    ----------
    doc
        A single document
    Returns
    -------
    a string with no punctuation
    """

    return doc.translate(str.maketrans('', '', string.punctuation))


def lemmatize(doc):
    """
    This function utilizes the lemmatizer in the spacy package to lemmatize all the words in the list

    Parameters
    ----------
    doc
        A single document

    Returns
    -------
    a string with lemmatized words

    """

    if not isinstance(doc, str):
        raise TypeError("row_string must be a string")

    filt_combined = []
    for word in nlp(doc):
        if word.lemma_ != '-PRON-':
            filt_combined.append(word.lemma_)

    new_df = " ".join(filt_combined)

    return new_df


def remove_stopwords(doc: str, allow_abbrev: bool = True):
    """
    This function utilizes the stopwords in the nltk package to remove the stopwords from the string.
    Further steps were taken that remove words commonly found in scientific articles.

    Parameters
    ----------
    doc
        A single document
    allow_abbrev
        A boolean indicating whether abbreviations should be considered stopwords. If true strings with character length of 2 or less are removed.

    Returns
    -------
    A string with stopwords removed
    """

    if not isinstance(doc, str):
        raise TypeError("row_string must be a string")

    if not isinstance(allow_abbrev, bool):
        raise TypeError("allow_abbrev must be a boolean")

    filt_combined = []

    for word in word_tokenize(doc):
        contains_letter = re.search('[a-zA-Z]', word) is not None
        stopword_check = word.lower() not in NLTK_STOPWORDS
        berteley_stopword_check = word.lower() not in STOPWORDS

        if contains_letter and stopword_check and berteley_stopword_check:
            if not allow_abbrev:
                if len(word.lower()) > 2:
                    filt_combined.append(word)

            else:  # allowing abbreviations
                filt_combined.append(word)

    filtered_ip = " ".join(filt_combined)

    return filtered_ip


def remove_html(doc: str):
    """
    Removes html tags from string
    Parameters
    ----------
    doc
        A single document

    Returns
    -------
    A string with html tags removed
    """
    soup = BeautifulSoup(doc, "html.parser")
    stripped_text = soup.get_text(separator=" ")
    return stripped_text


def remove_extraspace(doc: str):
    """
    Removes excess whitespace from a string
    Parameters
    ----------
    doc
        A single document

    Returns
    -------
    A string with excess whitespace removed
    """
    return ' '.join(doc.split())


def expand_contractions(doc: str):
    """
    Expands contractions using the contractions library

    Parameters
    ----------
    doc
        A single document

    Returns
    -------
    A string with contractions expanded

    Examples
    won't -> will not
    """
    return contractions.fix(doc)


def _identity(doc):
    return doc


def _map_docs(docs, func, filter=lambda doc: True):
    return [func(doc) for doc in docs if filter(doc)]


def preprocess(docs: List[str], allow_abbrev: bool = True, show_progress:bool=False) -> List[str]:
    """
    Wrapper function for all the preprocessing steps

    Parameters
    ----------
    docs
        A list of all the documents
    allow_abbrev
        Boolean indicating whether abbreviations should be allowed. If set to false all strings with length 2 or less will be removed
    show_progress
        If True, shows a progress bar for completion status

    Returns
    -------
    A list of strings that have been preprocessed
    """

    cleaned_docs = docs.copy()

    steps = [('Removing empty strings...', _identity, len),
             ('Removing all html tags...', remove_html),
             ('Expanding contractions...', expand_contractions), # don't -> do not
             ('Converting to lower case...', str.lower),
             ('Removing all punctuation...', remove_punctuation), # make sure hyphenated words are properly combined and grammatical hyphens are spaced appropriately
             ('Removing extra whitespace...', remove_extraspace),
             ('Removing empty strings...', _identity, len),
             ('Lemmatizing...', lemmatize),
             ('Removing stopwords...', partial(remove_stopwords, allow_abbrev=allow_abbrev)),
             ('Removing short strings', _identity, lambda doc: len(doc.split()) > 10)
             ]

    iterator = alive_it if show_progress else iter
    bar = iterator(steps)
    for title, func, *filter in bar:
        if show_progress:
            bar.title(title)
        cleaned_docs = _map_docs(cleaned_docs, func, **{'filter': filter[0]} if filter else {})

    return cleaned_docs


@contextlib.contextmanager
def alive_bar_joblib(bar):
    """Context manager to patch joblib to report into progress bar given as argument"""
    class AliveBatchCompletionCallback(parallel.BatchCompletionCallBack):
        def __call__(self, *args, **kwargs):
            if bar and not isinstance(bar, contextlib.nullcontext):
                for i in range(self.batch_size):
                    bar()
            return super().__call__(*args, **kwargs)

    old_batch_callback = parallel.BatchCompletionCallBack
    parallel.BatchCompletionCallBack = AliveBatchCompletionCallback
    try:
        yield bar
    finally:
        parallel.BatchCompletionCallBack = old_batch_callback


def preprocess_parallel(docs: List[str], n_workers: int = 4, allow_abbrev: bool = True, show_progress: bool=True) -> List[str]:
    """
    Parallelizes the preprocessing by splitting the documents evenly amongst the n_workers

    Parameters
    ----------
    docs
        List of documents
    n_workers
        Number of workers to be assigned preprocessing in joblib

    allow_abbrev
        Boolean indicating whether abbreviations should be allowed. If set to false all strings with length
        2 or less will be removed

    show_progress: bool
        If True, shows a progress bar for completion status

    Returns
    -------
    A list of strings that have been preprocessed

    """
    # remove all empty strings
    docs = list(filter(len, docs))

    with alive_bar(len(docs)) if show_progress else contextlib.nullcontext() as bar:
        with alive_bar_joblib(bar):
            clean_docs = Parallel(n_jobs=n_workers)(delayed(preprocess)([doc], allow_abbrev, show_progress=False)
                                                for doc in docs)
    # flatten the list of lists into a single list
    clean_docs = [item for sublist in clean_docs for item in sublist]

    return clean_docs
